/****************************************************************************
 * TP d'info-528 pour décoder les QRCodes (version 1)
 *
 * Références :
 *   - http://en.wikipedia.org/wiki/QRcode
 *   - http://en.wikiversity.org/wiki/Reed%E2%80%93Solomon_codes_for_coders
 *   - http://www.thonky.com/qr-code-tutorial/
 *
 ****************************************************************************/

#ifndef _QRCODES_H
#define _QRCODES_H


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <pbm.h>  /* libnetpbm : images au format pbm */


/* un type pour les octets (8 bits) */
typedef unsigned char byte;

/* largeur et hauteur des QR codes version 1 */
#define SIZE_V1 21

/* ATTENTION : un QR code est représenté par un tableau à deux dimensions 
 * qr[colone][ligne].
 * C'est inversé par rapport à la notation mathématique habituelle qui met
 * l'abscisse en premier... */
typedef bit **QRCode;
QRCode alloc_qrcode(int size);
void free_qrcode(QRCode qr, int size);


/* une macro pour obtenir le i-ème bit d'un octet (0 = bit de poids faible)
 * (Note : fonctionne aussi pour les entiers.)*/
#define BIT(u,i)            (((u)>>(i)) & 1)

/* fonctions d'affichage diverses et variées (utile pour les tests) */
void print_bits_array(const bit *T, int t);
void print_bits_polynomial(const bit *P, int d);
void print_bytes_polynomial(const byte *P, int d);
void print_byte(byte c);
void print_byte_matrix(int height, int width, byte **M);



/*************************************
 * lecture / écriture de fichier PBM *
 *************************************/

/* lire un fichier PBM et obtenir un QR code brut
 *   - image est le nom du fichier
 *   - margin est la taille, en pixel, de la marge
 *   - module_size est la taille, en pixel, des modules */
void pbm_to_qrcode_version1(const char *image, int margin, int module_size, QRCode qr);

/* écrire un fichier PBM binaire et obtenir un QR code brut
 *   - image est le nom du fichier
 *   - module_size est la taille, en pixel, des modules */
void qrcode_version1_to_pbm(const char *image, int module_size, QRCode qr);



/******************************************
 * parties fixes d'un QR code (version 1) *
 ******************************************/
/* renvoie le nombre de pixels fixes incorrects */
int check_fixed_bits_version1(QRCode qr);


/*********************
 * format du QR code *
 *********************/

/* récupère les 15 bits de format, aux deux endroits
 * sans appliquer le masque */
void get_raw_format_bits(QRCode qr, bit format[2][15]);

/* récupère les 15 bits de format, aux deux endroits
 * en appliquant le masque */
void get_format_bits(QRCode qr, bit format[2][15]);

/* fonction finale qui récupère les 5 bits de formats.
 * S'il avait des erreurs qu'on n'a pas pu corriger, la valeur de retour est
 * -1, sinon c'est 0. */
/* Pour le TP1, il suffit de vérifier si les deux formats sont identiques. */
int get_format(QRCode qr, int *correction_level, int *mask);



/**********************************
 * lecture des données du QR code *
 **********************************/

/* nombre d'octets utilisés par les données
 * (incluant la correction d'erreurs)*/
#define NB_BYTES 26

/* modes de codage */
#define NUMERIC_MODE 1
#define ALPHANUMERIC_MODE 2
#define BYTE_MODE 4
#define END_MODE 0

/* autres modes, non utilisés pour les TP */
#define FNC1_1 5
#define FNC1_2 9
#define ECI_MODE 7
#define KANJI_MODE 8
#define APPEND_MODE 3


/* applique le masque numéro m au qrcode : voir
 * http://en.wikiversity.org/wiki/Reed%E2%80%93Solomon_codes_for_coders#Masking
 * pour la liste des masques possibles */
void data_mask(QRCode qr, int mask);

/* récupère un octet de données à partir des coordonnées du coin haut/gauche
 * du rectangle correspondant */
byte get_upward_byte(QRCode qr, int i, int j);
byte get_downward_byte(QRCode qr, int i, int j);

/* récupère tous les octets de données */
void get_data_bytes(QRCode qr, byte *data);



/************
 * décodage *
 ************/

/* récupère un bit dans le tableau des octets de données */
bit get_bit(const byte *data, int nb) ;

/* récupère plusieurs bits consécutifs dans le tableau des octets de données
 * Le résultat est renvoyé dans un entier. Il ne faut donc pas demander trop
 * de bits en même temps ! */
unsigned int get_bits(const byte *data, int offset, int nb);

/* récupère les 4 bits du mode à un offset donné */
int get_mode(const byte *data, int offset);

/* récupère le nombre de symboles suivant un mode */
int get_length(const byte *data, int offset, int mode);

/* les fonctions de décodages des différents modes. La chaine result
 * doit déjà être allouée...
 * Elles renvoient le nombre de bits qu'elles ont lus dans le tableau des
 * données.*/
 int decode_bytes(const byte *data, int offset, int nb, char *result);
 int decode_numeric(const byte *data, int offset, int nb, char *result);
 int decode_alphanumeric(const byte *data, int offset, int nb, char *result);

/* décode et affiche le contenu du QR code contenu dans un fichier */
void decode_file(const char *filename, int margin_size, int module_size);

/* affiche toutes les informations importantes du QR code contenu dans un
 * fichier */
void info_file(const char *filename, int margin_size, int module_size);



/**********************************
 * correction d'erreurs du format *
***********************************/

/* calcule le syndrome du format */
void format_syndrome(const bit F[15], bit S[10]);

/* encode un format (5 bits), sans appliquer le masque */
void encode_format(const bit format_data[5], bit F[15]);

/* corrige le format (15 bits) en cherchant parmis tous les formats valides,
 * celui qui est le plus proche du format donné en argument.
 * Si c'est possible, le format est modifié, et la valeur de retour est égale
 * au nombre de bits modifiés.
 * Si ce n'est pas possible, le format n'est pas modifié, et la valeur de
 * retour est -1. */
int correct_format(bit F[15]);



/************************************
 * Correction d'erreurs des données *
 ************************************/

/* niveaux de correction d'erreurs */
#define ERROR_LEVEL_L 1
#define ERROR_LEVEL_M 0
#define ERROR_LEVEL_Q 3
#define ERROR_LEVEL_H 2
/* nombre d'octets de données pour chacun des niveaux de correction d'erreurs */
/* initialisation dans le fichier main.c, avant la fonction main */
int NB_DATA_BYTES[4] ;

/* initialisation des table de log et puissances */
/* appelée dans le fichier main.c, au début de la fonction main */
void init_tables(void);

/* opérations sur les coefficients des polynomes (octets) */
byte mult_coeff(byte a, byte b);
byte div_coeff(byte a, byte b);
byte inv_coeff(byte a);
byte pow_coeff(byte a, int n);


/* pivot de Gauss, en place, pour résoudre un système d'équations linéaires
 * S'il y a une solution unique, elle est mise dans dans "X" et la fonction
 * renvoie 0. Sinon, elle renvoie -1. */
int solve_linear_system(int height, int width, byte **M, byte *result);

/* évaluation du polynome P, de degré d, sur la valeur x. */
byte eval_poly(const byte *P, int d, byte x);

/* calcul des syndromes pour les octets de données */
void data_syndrome(const byte *data, int n, byte *S);

/* calcul des coefficients du polynome localisateur d'erreurs.
 *   - S contient les syndromes
 *   - le résultat (coefficients du polynome localisateur) est mis dans L
 * La valeur de retour correspond au nombre d'erreurs trouvées. */
int error_locator(int nb_correction_bytes, const byte *S, byte *L);


/* cherche la localisation des erreurs en cherchant les zeros du polynome de
 * localisation.
 *   - L contient le polynome de localisation
 *   - le résultat est stocké dans X
 * La valeur de retour est le nombre d'erreurs localisées. */
int find_locations(int nb_errors, const byte *L, byte *X);


/* calcul des valeurs d'erreurs en résolvant un système linéaire
 *   - S contient les syndromes
 *   - X contient les localisations des erreurs
 *   - le résultat est mis dans V
 * La valeur de retour est 0 si la résolution a donné une unique solution, et
 * -1 sinon. */
int error_values(int nb_correction_bytes, int nb_errors, const byte *S, const byte *X, byte *V);


/* corrige (si c'est possible) les octets de données */
int correct_data(byte *data, int correction_level);


/*********
 * tests *
 *********/

/* la fonction de test. */
void do_tests(void);


#endif   /* _QRCODES_H */


/* vim: set tabstop=2 shiftwidth=2 : */
